//
//									VEST
//   Stacking Velocity Estimation for 2D-3D CMP gathers w/o NMO corrections
//   stored in SEG-Y (IEEE) format with a native-byte-order 
//
// Copyright 2011-2015 by Yuri Ganshin, CMI, University of Wyoming, Laramie.
// http://www.uwyo.edu/cmi/
//
// This program is free software which means that the user have the freedom
// to run, copy, distribute, study, modify and improve the software.
//
//===========================================================================
//    Compile with: gcc -Wall vest.c -lm -O3 -o vest
//===========================================================================
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <fcntl.h>
#include <sys/types.h>
#include <math.h>
#include <float.h>  // declares FLT_MAX
/////////////////////////////////////////////////////////////////////////////
#define ENDCMP	 99999999	// Some impossible number to indicate END of cycle
#define LOW_FOLD 6			// Lowest CMP Fold to run calculations
#define MAXFOLD 200			// Maximum possible fold in CMP gather
#define VNUM	25			// Number of trial velocities within the corridor
//
#ifndef FLT_MAX
#define FLT_MAX 1e+37
#endif	//  Tells CPP to define FLT_MAX only if it isn't defined already 

#ifndef RAND_MAX
#define RAND_MAX 0x7fffffff
#endif

int main(argc,argv)
int argc;
char **argv;
{
	FILE *fp,*segin,*sego1,*sego2;
	int TIME_STA,TIME_FIN; // To make sure analysis window is within the limits	
	int cdp,cdpcur,cdplast,units,count,maxvi;
	int i,j,t,n,vi,tr,ns,FOLD,HYPER=1;
	int rwin,win,halfwin,res;	
	int *hdrinp,*hdrout;	
	short *hdrs;
		
	long file_size,foffset,count1=0,count2=0;
	unsigned int tr_size,traces,cdps;
          
	char cbuf[200],outvel[200],outcoh[200],fhdr[3600];
	
	float si,vg,t0,v;
	float dv,vsta,vfin,corr_w;
	float *buf[MAXFOLD],*Vs,*Vsmo,*coher,*bufg;
	float *dep,offs1[MAXFOLD];
	
	double offs2[MAXFOLD],offs4[MAXFOLD],*sum1,*sum2,*gw,*Vq;
	double sir,tmp,ftime,v2,v4,v8,t02,ampl,sembl,maxsem,C3;
	// for Parabolic Interpolation:
	double vels[VNUM],sems[VNUM],mincoh;
	double Va,Vb,Vc,Sa,Sb,Sc,dV1,dV2,dS1,dS2,dS3,vel;
	double val1,val2;
	// parameters for analytic power-law expression  V = v0 + alfa*t^beta
	double v0,alfa,beta;
	
	time_t system_time;
	float time_smo();
	
	/**
	 Conventionally, normal moveout is characterized by the following traveltime
	 equation valid for the homogeneous earth assumption with planar interface:
	 t^2 = t0^2 + x^2/Vs^2		[hyperbola equation]
	 At every CMP gather and every time sample, Velocity Spectra is calculated
	 for a set of trial stacking velocities [Vs] using semblance as coherency
	 measure. The semblance value reflects how well the moveout path corresponding
	 to the trial NMO velocity fits the actual moveout of reflected signal in the
	 data.
	 In the code, velocity spectra is estimated with a relatively few number
	 of trial velocities (VNUM=25) and is fitted to a Parabolic curve. Inverse
	 parabolic interpolation enables accurate estimate of the peak semblance 
	 value and the corresponding velocity value.
	 
	 The semblance as defined by Neidell and Taner [1971] is given by equation:
	 S = ∑[∑a(t,x)]^2/(N∑∑a(t,x)^2)
	 with the inner sums over CMP traces with variable offsets 'x', while the
	 outer sums correspond to a time-smoothing with WIN-length operator centered
	 at time index 't'. Semblance is maximized when the values a(t,x) do NOT
	 vary with trace index 'x'. Time-smoothing is necessary to increase semblance
	 stability for coherent events located close to zero-crossings.
	 Semblance with weights equation:   
	 S = ∑w(t)[∑a(t,x)]^2/(N∑w(t)∑a(t,x)^2)
	 Time smoothing is represented in the code with the Gaussian function.
	 Big smoothing window allows to bypass areas with poor reflectivity.
	 
	 Amplitude values a(t,x) can be taken at a discrete (integer) time samples
	 and at a decimal ones using cubic spline interpolation. In the last case,
	 the processing speed increases several times without visible gain in
	 accuracy and resolution.  [ tested by the author ]
	 **/	
	if(argc!=4)
	{
		printf("==========================================================================\n");
		printf("|     Detailed Stacking Velocity Estimation for 2D-3D CMP gathers        |\n");
		printf("|     based on conventional semblance [Neidell and Taner, 1971]          |\n");
		printf("|     and inverse parabolic interpolation for the peak value location    |\n");
		printf("|------------------------------------------------------------------------|\n");
		printf("|       Utilizes non-hyperbolic reflection traveltime approximation      |\n");
		printf("|       for isotropic layered media [three-term equation]                |\n");
		printf("|       described in Taner and Koehler [1969] and Alkhalifah [1997]      |\n");
		printf("==========================================================================\n");
		printf("|                PARAMETERS DESCRIPTION                                  |\n");
		printf("| SEG_inp = CMP-sorted SEGY-IEEE file with a native-byte-order           |\n");
		printf("| SEG_out = Output Velocity [SEG_out.vel] and Coherency [SEG_out.coh]    |\n");
		printf("| PARAM_inp = Parameter Text file with the following parameters:         |\n");
		printf("|------------------------------------------------------------------------|\n");
		printf("|    V0, ALPHA, and BETHA are parameters describing the power-law        |\n");
		printf("|        guide velocity function [use VELAN routine for its estimation]  |\n");
		printf("|    CORR_W - Velocity Corridor Width relative to the guide velocity     |\n");
		printf("|        function [estimated with VELAN routine; usually 0.1 - 0.3]      |\n");
		printf("|    WIN - window for smoothing the semblance estimates, usually is set  |\n");
		printf("|        at about half the dominant period of the wavelet [ 12-24 ms ]   |\n");
		printf("|    MINCOH  - Minimum coherency [0.2 - 0.5] for velocity assignment     |\n");
		printf("|    HYPER - 1 for hyperbolic traveltime, else NON-hyperbolic curve      |\n");
		printf("|------------------------------------------------------------------------|\n");
		printf("| The parameter file should contain three lines, with the first two      |\n");
		printf("| reserved for comments, and the last one filled with space-separated    |\n");
		printf("| seven values as in the example below:                                  |\n");
		printf("|------------------------------------------------------------------------|\n");
		printf("| #                  Sample Parameter File                               |\n");
		printf("| #  V0    ALPHA   BETHA  Corr_W  WIN   MINCOH  HYPER                    |\n");
		printf("|  2100.0  1350.0  0.60    0.15   20     0.25    0                       |\n");
		printf("==========================================================================\n");
		printf("USAGE: %s SEG_inp SEG_out PARAM_inp\n\n",argv[0]);
		return 1;
	}
	const char* inpname = argv[1];
	const char* outname = argv[2];
	const char* parname = argv[3];
	/*************************************************************************/

	strcpy(outvel,outname);  strcat(outvel,".vel");
	strcpy(outcoh,outname);  strcat(outcoh,".coh");
		
	//system("date");
	if(sizeof(int) !=4 ) { printf("\nUnsupported INT type\n"); return 1; }
	if(sizeof(float) !=4 ) { printf("\nUnsupported FLOAT type\n"); return 1; }
	
   	hdrinp=(int*) malloc(240);		// Actually, sixty 4-byte words
   	hdrout=(int*) malloc(240);	
	
	/*************************************************************************/
	/***********		Read    P A R A M E T E R   file     *******************/
	/*************************************************************************/
   
   if( (fp=fopen(parname,"r"))==NULL)	// Plain TXT file
   { printf("File \"%s\" opening ERROR\n",parname); return 1; }
   
   count=0; while(fgets(cbuf,150,fp)) count++;
	if(count != 3) 
   { 
		printf("\n  ERROR: the PARAMETER file should contain 3 strings,\n");
		printf("  with the first two containing comments,\n");
		printf("  and the last one filled with space-separated values\n");			 
		return 1; 
	}

	res=sscanf(cbuf,"%lf%lf%lf%f%d%lf%d\n",
				  &v0,&alfa,&beta,&corr_w,&rwin,&mincoh,&HYPER);
	if(res != 7)
   { 
		printf("\n  ERROR: the PARAMETER file should contain 7 values\n");
		return 1;
	}
	printf("\n  Processing PARAMETERS:\n");
	printf("  V0 = %.0lf  Alpha = %.0lf  Betha = %.2lf\n",v0,alfa,beta);
	printf("  Relative Velocity Corridor Width = %.2f\n",corr_w);
	printf("  Smoothing Window = %d\n  Minimum Coherency = %.2lf\n",rwin,mincoh);
	if(HYPER==1) printf("  Hyperbolic moveout corections will be applied\n");
	else printf("  Non-hyperbolic moveout corections will be applied\n");			

	/*************************************************************************/
	/*********************  Reading SEGY file  *******************************/
	/*************************************************************************/
        
	if((segin = fopen(inpname,"rb")) == NULL)		// BINARY mode for opening!
	{ printf("\nERROR:  cannot find file %s\n\n",inpname); return 1; }
  
	fseek(segin, 0L, SEEK_END);	// go to file's end
	file_size = ftell(segin);
	fseek(segin, 0L, SEEK_SET);	// go to the beginning	

	/*******************************************************************/
	/********       READ Textual and Binary File Headers        ********/
    
	fread (&fhdr, 3600, 1, segin);      // Read 1 object 3600-bytes long
	hdrs=(short*)fhdr;
	si = (float)hdrs[1608]/1000.0;      // SAMPLE INTERVAL, ms
	ns = hdrs[1610];                    // NUMBER of SAMPLES	
    
	printf("\nFile: '%s', SI=%.1f ms, NS=%d samples",inpname,si,ns); 	
	if(si<=0.0 || si>10.0 || ns<=0 || ns>25000) 
	{ printf("\nERROR:  unsupported file format !\n"); fclose(segin); return 1; }
    
	tr_size=(unsigned int)(ns*sizeof(float));    
	traces=(long)(file_size-3600)/(tr_size+240);
	printf(", traces = %d\n",traces);

	units = hdrs[1627];
	printf("Measuring system code = %d",units);
	if(units==1) printf(" --> meters\n");
	else if(units==2) printf(" --> feet\n");
		else printf("\nWARNING:  unknown units !!!\n");

	if(fread(hdrinp,240,1,segin) != 1)	// Read 1 object 240-bytes long
	{ printf("\nERROR reading first trace\n"); fclose(segin); return 1; }
	cdpcur=hdrinp[5];				// CMP # of the first trace

	for(i=0;i<60;i++) hdrout[i]=hdrinp[i];	
	
	foffset = 3600L + (traces-1)*(tr_size+240);
	fseek(segin,foffset,SEEK_SET);
	if(fread(hdrinp,240,1,segin) !=1 ) 
	{ printf("\nERROR reading last trace\n"); fclose(segin); return 1; }
	cdplast=hdrinp[5];			// CMP # of the last trace
	
	printf("\nFirst trace CMP# = %d, last CMP# = %d\n\n",cdpcur,cdplast);
	/**********************************************************************/
	/****************      Definitions and Memory       *******************/

	win=(int)(0.0001+(float)rwin/si);	// Full window length, samples
	halfwin=win/2;			// Half window length, samples
	win=halfwin*2+1;		// Window must be odd number to make it symmetrical
	
	TIME_STA=win;			//  To stay within array's limits
	TIME_FIN=ns-win-1;      //  To stay within array's limits
	
	sir=1000.0/si;			// Time samples per second
 
	for(FOLD=0;FOLD<MAXFOLD;FOLD++) buf[FOLD]=(float*)malloc(tr_size);
	
	Vs=(float*)malloc(tr_size);
	Vsmo=(float*)malloc(tr_size);
	dep=(float*)malloc(tr_size);	
	coher=(float*)malloc(tr_size);	
	bufg=(float*)malloc(tr_size);
	Vq=(double*)malloc(ns*sizeof(double));	 // Quartic Velocity correction

        // Semblance related 
	sum1=(double*)malloc(win*sizeof(double));		
	sum2=(double*)malloc(win*sizeof(double));
		
	//   Array of Gaussian  Weights for semblance calculation
	gw=(double*) malloc(win*sizeof(double)); 
	for(i=0;i<win;i++)   // win - full window size [samples]
	{
		ampl=(double)abs(halfwin-i); tmp=sqrt(M_PI)*ampl/(double)win;
		gw[i]=exp(-1.0*tmp*tmp);	// gw[i]=1.0 for i==halfwin
	}
		
	//   Seed Random Numbers Generator
	system_time = time(NULL);		// in seconds since January 1, 1970
	srand((unsigned)system_time);	// system time as a seed number

	/**********************************************************************/	
	printf("\n   Guide Velocity functions:\n");
	printf(" TIME, ms     Depth      Vs         Vint      Quartic, units/s\n");
	
	Sa=0.0;  dep[0]=0.0;	
	for(i=0;i<ns;i++)
	{
		ftime = (double)i/sir;		// time in 'sec'
		val1 = pow(ftime,beta);
		vel = v0 + alfa*val1;		// Fitted stacking Velocity Guide Function
		bufg[i] = (float)vel/sir;	// units/sample		
		
		val2 = pow(ftime*ftime,beta);
		// Lumley's Vint formula from the SEP Report #75, 1997
		Vc = sqrt(v0*v0+v0*2*alfa*(beta+1)*val1+(beta*2+1)*alfa*alfa*val2);
		Vc /= sir;						// units/sample
		
		if(i>0) dep[i] = dep[i-1] + Vc/2.0; 
		// one-way time during half sample interval
		
		Sa += Vc*Vc*Vc*Vc;
		Vq[i] = Sa/(i+1);  //  Quartic Velocity estimation taken from
		//	 Biondi, et al., "Nonstretch normal moveout ...", Geophys., 2014
		
		if(i%50==0)
			printf("%7.0f   %9.1f   %9.1f  %9.1lf  %9.1lf\n",
					 si*i,dep[i],bufg[i]*sir,Vc*sir,pow(Vq[i],0.25)*sir);
	}	
	printf("\n");

/**************************************************************************/
/********************   OUTPUT  FILES   ***********************************/

	if((sego1 = fopen(outvel,"wb")) == NULL)	// BINARY mode for output
	{ 
		printf("\nERROR:  cannot create file %s\n\n",outvel); 
		fclose(segin); return 1; }
	if((sego2 = fopen(outcoh,"wb")) == NULL)
	{ 
		printf("\nERROR:  cannot create file %s\n\n",outcoh); 
		fclose(segin); return 1; }
    
	fwrite (&fhdr,3600,1,sego1);   // Writes 1 object 3600 bytes long    
	fwrite (&fhdr,3600,1,sego2);

/***************************  Major Cycle   ******************************/
/*************************************************************************/
	
	FOLD=0; cdps=0;
	fseek(segin,3600L,SEEK_SET);
	
   while(1)
   {         
		if( fread(hdrinp,240,1,segin) == 1 ) 
		{
			hdrs=(short*)hdrinp;	// hdrs[14]=2 for dead traces
			if(hdrs[14] == 2) { fseek(segin,tr_size,SEEK_CUR); continue; }
			cdp=hdrinp[5];
		}
		else cdp=ENDCMP;  // To finish processing of the last accumulated CMP

      if(cdp==cdpcur)	// Accumulate traces of a CMP gather into the buf[]
      {
         fread(buf[FOLD],tr_size,1,segin); // read 1 object trace_sz long
         tmp=(double)(abs(hdrinp[9]));
         offs1[FOLD]=tmp;
         tmp *= tmp;
         offs2[FOLD]=tmp;
         offs4[FOLD]=tmp*tmp;			
         FOLD++;
         if(FOLD>MAXFOLD) { printf("MAXFOLD exceeded!\n"); break; }
      }
      else         //  Processing of accumulated CDPs 
      {
			if(FOLD<LOW_FOLD)  // No processing, return to trace accumulation
         {
            if(cdp==ENDCMP) 
            { printf("%6d \t\tLow-fold CMP# %d\n",cdps+1,cdplast); break; }
            cdps++;  cdpcur=cdp;  FOLD=0;
            printf("%6d \t\tLow-fold CMP# %d\n",cdps,cdpcur);
            for(i=0;i<60;i++) hdrout[i]=hdrinp[i];
            fseek(segin,-240L,SEEK_CUR);
            continue; 
         }
			for(n=0,t0=0.0;n<ns;n++,t0+=1.0)  /////////   Cycle by t0 and 'n'
			{  
				vg=bufg[n]; // use Guide function as default velocity value +
				Vs[n] = vg + vg*0.02*(1.0-2.0*rand()/RAND_MAX);  // 2% noise
				coher[n]=0.0;				// default coherence value == 0.0
				if(buf[0][n]==buf[0][n+1]) continue;
				
				sembl=maxsem=0.0; 
				t02=(double)t0*t0;
				
				vsta = vg*(1.0 - corr_w);  vfin = vg*(1.0 + corr_w);
				dv   = vg*2.0*corr_w/VNUM;	
				maxvi=0;
				for(vi=0,v=vsta;vi<VNUM;vi++,v+=dv)		// Velocity scanning
				{
					v2=v*v;  v4=v2*v2;  v8=v4*v4;
					
					if(HYPER) C3 = 0.0;  // hyperbolic moveout
					else C3=(v4-Vq[n])/(4.0*t02*v8); // third-term long-offset correction
													
					for(i=0;i<win;i++) sum1[i]=sum2[i]=0.0;
					count=0;
					for(tr=0;tr<FOLD;tr++)
					{	// Non-hyperbolic traveltime approximation with the
						// fourth-order equation described by Taner and Koehler (1969)
						
					   if(offs1[tr]>1.5*dep[n]) continue; // avoid long offsets
						
						val1 = t02 + offs2[tr]/v2;
						
						if(HYPER) ftime = sqrt(val1);
						else // Do NOT mix hyperbolic and non-hyperbolic traveltimes!
						{
							val2 = val1 + C3*offs4[tr];
							if(val2<0.0) continue;		
							ftime = sqrt(val2);
						}

						t=(int)ftime;
						
						if(t<TIME_STA || t>=TIME_FIN) continue;
					   j=t-halfwin;
						for(i=0;i<win;i++)
						{
							ampl=(double)buf[tr][j+i];
							sum1[i]+=ampl;			// inter-trace accumulation first
							sum2[i]+=(ampl*ampl);	// inter-trace accumulation first
						}
						count++;					
					}					
					if(count>=LOW_FOLD)
					{  
						tmp=ampl=0.0;
						for(i=0;i<win;i++) 
						{ sembl=sum1[i]; ampl+=(gw[i]*sembl*sembl); tmp+=(gw[i]*sum2[i]);}
						if(tmp>0.0) sembl=ampl/(tmp*count); // conventional semblance
						else sembl=0.0;
					}
					else sembl = 0.0;
					
					vels[vi] = v;  sems[vi] = sembl;
					if(sembl>maxsem) { maxsem=sembl; maxvi=vi; }
				}												// Velocity scanning

				if(n>ns/5 && n<4*ns/5)
				{
					if(maxvi>1 && maxvi<VNUM-2) count1++;
					else count2++;
				}		
				
		/****    Calculate interpolated Semblance and corresponding Velocity    ****/
		// The Parabolic Interpolator fits parabola through the highest three samples
		// of a global peak to estimate the true peak location and height 
		 				
				if(maxvi<=0) { vel = vsta;  sembl = maxsem; }
				else if(maxvi<VNUM-1)
				{
					Va=vels[maxvi-1]; Vb=vels[maxvi]; Vc=vels[maxvi+1];
					Sa=sems[maxvi-1]; Sb=sems[maxvi]; Sc=sems[maxvi+1];
					dV1 = Vb-Va;  dV2 = Vb-Vc;
					dS1 = Sb-Sa;  dS2 = Sb-Sc;  dS3 = Sc-Sa;
					
					vel = Vb - 0.5*(dV1*dV1*dS2-dV2*dV2*dS1)/(dV1*dS2-dV2*dS1);
					sembl = Sb + 0.25*dS3*dS3/(2.0*Sb-Sc-Sa);
				}
				else { vel = vfin;  sembl = maxsem; }				

				if(sembl>1.0) coher[n] = 100.0;
				else coher[n] = sembl*100.0;
								
				if(sembl>mincoh) Vs[n] = vel; 
				
			}				////////////   Cycle by t0 and 'n'   ////////////
         //////////////////////////////////////////////////////////////////////

			for(n=0;n<ns;n++) Vs[n] *= sir;

			// Smooth Velocity function
		   for(n=0;n<ns;n++) Vsmo[n] = time_smo(n,ns,win,gw,Vs);
   
			fwrite(hdrout,240,1,sego1); 
			if(fwrite(Vsmo,tr_size,1,sego1) != 1) 
			{ printf("\nERROR writing to velocity file!\n"); break; }
			fwrite(hdrout,240,1,sego2); 
			if(fwrite(coher,tr_size,1,sego2) != 1)          
			{ printf("\nERROR writing to coherency file!\n"); break; }
					
			cdps++;
		   if(cdpcur%20==0) printf("%6d \t\tCMP# %d \tFOLD %d\n",cdps,cdpcur,FOLD);
				
			if(cdp == ENDCMP) { printf("\nLast CMP processed: %d\n",cdpcur); break; }
			else 
			{
				cdpcur=cdp;  FOLD=0; 
				for(i=0;i<60;i++) hdrout[i]=hdrinp[i];			
				fseek(segin,-240L,SEEK_CUR);
			}					  
		}				//  Processing of accumulated CMP traces 		
   }  
  
   fclose(segin); fclose(sego1); fclose(sego2);
	
	printf("\n");
	printf("Velocity Corridor was too narrow in %.4f percent of cases\n",
						(float)count2/count1);
   
   //system("date");
   printf("\nDone\n");
   return 0;
}
/***********************************************************************/
float time_smo(n,ns,win,W,Y)  /* Smoothes array Y[0..ns] at sample 'n' */
float *Y;
int n,ns,win;    /* using smothing operator given with array W[0..win] */
double *W;       /*   W[0] correspond to the sample NEXT to central    */
{
   int k,_j,j_;
   double weight,Y_sum,w_sum;
   
   Y_sum=Y[n]; w_sum=1.0;  /** Central value **/
   for(k=1;k<=win;k++)
   {		
      weight=W[k-1];	  	
		j_=n+k;
		if(j_<ns) { Y_sum+=Y[j_]*weight; w_sum+=weight; }
		_j=n-k;
		if(_j>=0) { Y_sum+=Y[_j]*weight; w_sum+=weight; }		 
   }		
   return((float)(Y_sum/w_sum));   
}
